/*
 * CONVGATAN - 	Convert GATAN files into MRC format on a UNIX platform
 *
 *		V2.0 august '93 J. Brink
 *
 *		W.M. Keck Center for Computational Biology
 *		Department of Biochemistry
 *		Baylor College of Medicine
 *		1 Baylor Plaza
 *		Houston, TX 77030, U.S.A.
 *		(713) 798 - 6989
 *
 */
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <limits.h>
#include "gatan.h"
#include "mrc.h"


main(argc, argv)
int argc;
char** argv;
{
	FILE *f1, *f2;
	int i;
	time_t systim;
	char *tim;
	long  int datasize;
	struct gatan gatanh;
	struct mrc mrch;
	float dmin, dmax, dmean, ampl;
	typedef struct COMPLEX {
	  float r,i;
	} COMPLEX;

	union {
	  unsigned char *b;
	  short int *s;
	  long *l;
	  float *f;
	  COMPLEX *c;
	  void *v;
	} EDPdata;

	if (argc !=3)	{
	    printf("Usage: convgatan2 infile outfile\n");
	    exit(1);
	    }
/*
 *
 *      Open the GATAN file for inpt and the MRC file for output
 *
 */
	if (( f1 = fopen(argv[1], "r")) == NULL)	{ 
	     fprintf(stderr,"Input file did not exist !\n");
	     exit(1);
	   }

	if ((f2 = fopen(argv[2], "w")) == NULL)         {
	     fprintf(stderr,"Couldn't create output file !\n");
	     exit(1);
	   }
/*
 *
 *      Read the GATAN header to figure out the number of pixels
 *      along X and Y, the # of bytes/pixel and the data type.
 *
 */
	i = fread(&gatanh, sizeof(gatanh), (size_t)1, f1);
	if (ferror(f1)) {
	     fprintf(stderr,"Error reading GATAN header\n");
	     exit(1);
	   }
	if (feof(f1)) {
	  fprintf(stderr,"EOF on GATAN header encountered\n");
	  exit(1);
	}

	if (gatanh.ntype == 5) {
	  printf("Encountered packed complex data. Not handled by CONVGATAN2.\n");
	  exit(1);
	}
	if (gatanh.ntype == 7) {
	  printf("Encountered long integer data. Not handled by CONVGATAN2.\n");
	  exit(1);
	}
	
	printf("NX, NY of GATAN data : %i x %i\n",gatanh.nx,gatanh.ny);
	printf("# of bytes/pixel     : %i\n",gatanh.nbytes);
/*
 *
 *      Prepare to allocate the memory needed to read the GATAN data in
 *
 */
	datasize = (long) (gatanh.nx * gatanh.ny);

	if ((EDPdata.v = calloc ( (size_t)datasize,(size_t)gatanh.nbytes)) == NULL ) {
	      printf("Couldn't allocate memory.\n");
	      exit(1);
	    }
/*
 *
 *      Read the GATAN data into the allocated memory in the largest possible
 *      block size
 *
 */
	i = fread(EDPdata.v, (size_t)datasize, (size_t)gatanh.nbytes, f1);
	if (ferror(f1)) {
	     fprintf(stderr,"Error reading GATAN data.\n");
	     exit(1);
	   }
	if (feof(f1)) {
	  fprintf(stderr,"EOF on GATAN data encountered.\n");
	  exit(1);
	}
/*
 *
 *      Go through memory to compute the minimum, maximum and mean density values
 *      here.
 *      Also, make sure that we confirm to the MRC notation for the various data
 *      formats.
 */

	switch (gatanh.nbytes) {

/*
 *	Byte data (6)
 */
	case 1:
	  if (gatanh.ntype != 6) {
	    printf("Read wrong byte data type identifier from Gatan header: %i.\n",
		   gatanh.ntype);
	    exit(1);
	  }
	  mrch.mode = 0;
	  dmin = 255;
	  dmax = 0;
	  dmean = 0;
	  
	  for (i=1; i < datasize; i++) {
	    if ( EDPdata.b[i] < dmin )
	      dmin = EDPdata.b[i];
	    if ( EDPdata.b[i] > dmax )
	      dmax = EDPdata.b[i];
	    dmean = dmean + EDPdata.b[i];
	  }
	  break;

/*
 *	Signed 2 byte integer (short) (1)
 */
	case 2:
	  if (gatanh.ntype != 1) {
	    printf("Read wrong short data type identifier from Gatan header: %i.\n",
		   gatanh.ntype);
	    exit(1);
	  }
	  mrch.mode = 1;
	  dmin = SHRT_MAX;
	  dmax = -SHRT_MAX;
	  dmean = 0;
	  
	  for (i=1; i < datasize; i++) {
	    if ( EDPdata.s[i] < dmin )
	      dmin = EDPdata.s[i];
	    if ( EDPdata.s[i] > dmax )
	      dmax = EDPdata.s[i];
	    dmean = dmean + EDPdata.s[i];
	  }
	  break;

/*
 *	Either floating point IEEE standard 754 (2)
 *	 or
 *	Signed 4 byte integer (long) (7)
 */
	case 4:
	  if (gatanh.ntype != 2 && gatanh.ntype != 7) {
	    printf("Read wrong 4 byte data type identifier from Gatan header: %i.\n",
		   gatanh.ntype);
	    exit(1);
	  }
/*
 *	Floating point number IEEE standard 754 (2)
 */
	if (gatanh.ntype == 2) {
	    mrch.mode = 2;
	    dmin = FLT_MAX;
	    dmax = FLT_MAX;
	    dmean = 0;
	    
	    for (i=1; i < datasize; i++) {
	      if ( EDPdata.f[i] < dmin )
		dmin = EDPdata.f[i];
	      if ( EDPdata.f[i] > dmax )
		dmax = EDPdata.f[i];
	      dmean = dmean + EDPdata.f[i];
	    }
	  }
/*
 *	Signed 4 byte integer (long) (7)
 */
	  if (gatanh.ntype == 7) {
	    mrch.mode = 1;
	    dmin = LONG_MAX;
	    dmax = -LONG_MAX;
	    dmean = 0;
	    
	    for (i=1; i < datasize; i++) {
	      if ( EDPdata.l[i] < dmin )
		dmin = EDPdata.l[i];
	      if ( EDPdata.l[i] > dmax )
		dmax = EDPdata.l[i];
	      dmean = dmean + EDPdata.l[i];
	    }
	  }
	  break;

/*
 *	Complex data 2 floating point numbers (3)
 */
	case 8:
	  if (gatanh.ntype != 3) {
	    printf("Read wrong comple data type identifier from Gatan header: %i.\n",
		   gatanh.ntype);
	    exit(1);
	  }
	  mrch.mode = 4;
          dmin = FLT_MAX;
	  dmax = -FLT_MAX;
	  dmean = 0;


	  for (i=1; i < datasize; i++) {
            ampl = sqrt( EDPdata.c[i].r * EDPdata.c[i].r +
                        EDPdata.c[i].i * EDPdata.c[i].i );
            if ( ampl < dmin )
              dmin = ampl;
            if ( ampl > dmax )
              dmax = ampl;
	    dmean = dmean + ampl;
          }
	break;

	default: printf("Unknown GATAN mode encountered: %i.\n",gatanh.ntype);
	  exit(1);
	}
	
	dmean = dmean / datasize;
/*
 *
 *      Write the appropriate values into the MRC header variables
 *
 */
	mrch.nx = gatanh.nx;
	mrch.ny = gatanh.ny;
	mrch.nz = 1;


	mrch.nxstart = mrch.nystart = mrch.nzstart = 0;
	mrch.mx = gatanh.nx;
	mrch.my = gatanh.ny;
	mrch.mz = 1;
	mrch.xlength = gatanh.nx;
	mrch.ylength = gatanh.ny;
	mrch.zlength = 1.0;
	mrch.alpha = mrch.beta = mrch.gamma = 90.0;
	mrch.mapc = 1;
	mrch.mapr = 2;
	mrch.maps = 3;
	mrch.amin = dmin;
	mrch.amax = dmax;
	mrch.amean = dmean;
	mrch.ispg = 0;
	mrch.nsymbt = 0;
	for (i=0; i <29; mrch.extra[i++]=0);
	mrch.xorigin = 0;
	mrch.yorigin = 0;
	mrch.nlabl = 1;

/*
 *
 *      Assemble the first label for the new MRC header
 *
 */

	systim = time( (time_t *) 0);
	tim = ctime(&systim);
	tim[strlen(tim) - 1] = '\0';

	i = sprintf(mrch.label[0], "SSC data %i x %i, CONVGATAN2 on %s",
		gatanh.nx, gatanh.ny, tim);

/*
 *
 *      Fill all remaining labels with NULL's
 *
 */

	for (i=0; i<(80*9); *(mrch.label[1]+i++) = '\0');

/*
 *
 *      Write out the MRC header and next the MRC data
 *
 */

	i = fwrite(&mrch, sizeof(mrch), (size_t)1, f2);
	if (ferror(f2)) {
	     fprintf(stderr,"Error writing MRC header.\n");
	     exit(1);
	   }
	if (feof(f2)) {
	  fprintf(stderr,"EOF on MRC header encountered.\n");
	  exit(1);
	}
	  
	i = fwrite(EDPdata.v, (size_t)datasize, (size_t)gatanh.nbytes, f2);
	if (ferror(f2)) {
	     fprintf(stderr,"Error writing MRC data.\n");
	     exit(1);
	   }
	if (feof(f2)) {
	  fprintf(stderr,"EOF on MRC data encountered.\n");
	  exit(1);
	}

/*
 *      Make sure we free up the allocated memory !
 */
	free(EDPdata.v);

	fclose(f1);
	fclose(f2);
/*
 *      THE END
 */

      }











